[Setting name="Dot Size"]
int dotSize = 4;

class Block {
    string name;
    vec3 position;
    vec3 midpoint;
    bool iscustom;
    bool isblock = true;
    bool should_display = false;
    nat3[] blockUnits = {};
    string WaypointType = '';
    CGameCtnBlock::ECardinalDirections direction;
    bool isground = false;
    bool isghost = false;
    bool isfreeblock = false;

    Block (CGameCtnBlock@ gameBlock, uint saveCheck) {
        this.should_display = saveCheck >= 2;
        string idName = gameBlock.BlockModel.IdName;
        if (idName.ToLower().SubStr(idName.Length - 22) == ".block.gbx_customblock") {
            this.iscustom = true;
        } else if (idName.ToLower().SubStr(idName.Length - 10) == ".block.gbx") {
            this.iscustom = false;
        } else {
            this.iscustom = false;
        }
        
        this.midpoint = GetCtnBlockMidpoint(gameBlock);
        this.position = GetCtnBlockPosition(gameBlock);
        this.direction = gameBlock.Direction;
        this.name = gameBlock.BlockModel.Name;
        this.blockUnits.Reserve(gameBlock.BlockUnits.Length);
        for (uint i = 0; i < gameBlock.BlockUnits.Length; i++) {
            this.blockUnits.InsertLast(gameBlock.BlockUnits[i].AbsoluteOffset);
        }
        if (gameBlock.WaypointSpecialProperty !is null) {
            this.WaypointType = gameBlock.WaypointSpecialProperty.Tag;
        }
        this.isground = gameBlock.IsGround;
        this.isghost = gameBlock.IsGhostBlock();
        this.isfreeblock = IsBlockFree(gameBlock);
    }

    Block (CGameCtnAnchoredObject@ gameItem, uint saveCheck) {
        this.should_display = saveCheck >= 2;
        this.isblock = false;
        string idName = gameItem.ItemModel.IdName;
        if (idName.ToLower().SubStr(idName.Length - 9) == ".item.gbx") {
            this.iscustom = true;
        } else {
            this.iscustom = false;
        }
        this.position = gameItem.AbsolutePositionInMap;
        this.midpoint = this.position;
        this.name = gameItem.ItemModel.IdName;
        // this.blockUnits = {};
        if (gameItem.WaypointSpecialProperty !is null) {
            this.WaypointType = gameItem.WaypointSpecialProperty.Tag;
        }
    }

    bool differentHighlight() {
        switch (variantColor) {
            case VariantColorOptions::Disabled: return false;
            case VariantColorOptions::Ghost: return this.isghost;
            case VariantColorOptions::NotGhost: return !this.isghost;
            case VariantColorOptions::Ground: return this.isground;
            case VariantColorOptions::NotGround: return !this.isground;
            case VariantColorOptions::Free: return this.isfreeblock || !this.isblock;
            case VariantColorOptions::NotFree: return !this.isfreeblock && this.isblock;
            case VariantColorOptions::NotNormal: return this.isghost || this.isfreeblock || !this.isblock;
            case VariantColorOptions::Normal: return !(this.isghost || this.isfreeblock || !this.isblock);

            default: return false;
        }
    }

    void renderDot() {
        vec3 pos = Camera::ToScreen(this.midpoint);
        if (pos.z >= 0) {return;} // only happens when loading map

        float size;
        float distance = Math::Sqrt(-pos.z/32);
        size = (2*dotSize*(distance+3))/(distance*4);
        nvg::BeginPath();
        nvg::Circle(pos.xy, size);
        if (this.differentHighlight()) {
            if (this.isblock) {
                nvg::FillColor(vec4(1.f, 0.6f, 0.f, 0.9f));
            } else {
                nvg::FillColor(vec4(1.f, 0.4f, 0.f, 0.9f));
            }
        } else {
            nvg::FillColor(vec4(1.f, 0.f, 0.f, 0.9f));
        }
        nvg::Fill();
    }

    void renderDot_Debug() {
        vec3 pos = Camera::ToScreen(this.position);
        if (pos.z >= 0) {return;} // only happens when loading map

        float size;
        float distance = Math::Sqrt(-pos.z/32);
        size = (2*dotSize*(distance+3))/(distance*4);
        nvg::BeginPath();
        nvg::Circle(pos.xy, size);
        nvg::FillColor(vec4(0.f, 0.f, 1.f, 0.9f));
        nvg::Fill();
    }


    int opCmp(Block@ other) { // 1 is this lower, -1 is this higher
        if (this.iscustom && !other.iscustom) {
            return 1;
        } else if (!this.iscustom && other.iscustom) {
            return -1;
        }
        if (this.isblock && !other.isblock) {
            return -1;
        } else if (!this.isblock && other.isblock) {
            return 1;
        }
        if (this.name < other.name) {
            return -1;
        } if (this.name > other.name) {
            return 1;
        }
        return 0;
    }
}

// vec3 mapCoord(nat3 blockCoord) {
//     return vec3(blockCoord.x * 32 + 16, (blockCoord.y - 8) * 8 + 2, blockCoord.z * 32 + 16);}

// // const vec3 freeblockOffset = vec3(16, 2, 16);
// const vec3 freeblockOffset = vec3(16, 2, 16);
// vec3 offsetFreeblock(vec3 rot) {
//     // // I think do xrot, then zrot, then yrot
//     float yrot = rot.x;
//     // float xrot = rot.y;
//     // float zrot = rot.z;
//     vec3 offset = freeblockOffset;
//     float currentRot, currentLen;

//     // currentLen = Math::Sqrt(freeblockOffset.y**2 + freeblockOffset.z ** 2);
//     // currentRot = Math::Atan2(offset.y, offset.z) - xrot; // negative since game is in opposite direction
//     // offset.y = Math::Sin(currentRot) * currentLen;
//     // offset.z = Math::Cos(currentRot) * currentLen;

//     // currentLen = Math::Sqrt(freeblockOffset.y**2 + freeblockOffset.x ** 2);
//     // currentRot = Math::Atan2(offset.y, offset.x) + zrot; // negative since game is in opposite direction
//     // offset.y = Math::Sin(currentRot) * currentLen;
//     // offset.x = Math::Cos(currentRot) * currentLen;



//     currentRot = Math::Atan2(offset.x, offset.z);
//     currentLen = Math::Sqrt(offset.x ** 2 + offset.z ** 2);
//     currentRot += yrot;
//     offset.x = Math::Sin(currentRot) * currentLen;
//     offset.z = Math::Cos(currentRot) * currentLen;




//     // print('' + offset.x + ' ' + offset.y + ' ' + offset.z + ' ' + yrot + ' ' + xrot + ' ' + zrot);
//     return offset;

// }
