
bool blockUnitsEqual(nat3[]@ BlockUnits_A, nat3[]@ BlockUnits_B) {
    if (BlockUnits_A.Length != BlockUnits_B.Length) {
        return false;}
    for (uint i = 0; i < BlockUnits_A.Length; i++) {
        if (!nat3equals(BlockUnits_A[i], BlockUnits_B[i])) {
            return false;}
    }
    return true;}

bool nat3equals(nat3 a, nat3 b) {
    return (a.x == b.x && a.y == b.y && a.z == b.z);}

const float closeness = 0.01f;
bool vec3equals(vec3 a, vec3 b) {
    return (Math::Abs(a.x - b.x) < closeness && Math::Abs(a.y - b.y) < closeness && Math::Abs(a.z - b.z) < closeness);}

// if a road block has a custom block overlayed, it doesn't need to be changed
bool HasCustomOverlay(Block@ block, Block@[]@ blocks) {
    for (uint i = 0; i < blocks.Length; i++) {
        if (blocks[i].iscustom && blocks[i] !is block) {
            if (vec3equals(block.position, blocks[i].position) && (!checkOverlayHitbox || blockUnitsEqual(block.blockUnits, blocks[i].blockUnits) || block.WaypointType != '')) {
                return true;
            }
        }
    }
    return false;
}


bool HasCheckpointOverlay(Block@ block, Block@[]@ blocks) {
    for (uint i = 0; i < blocks.Length; i++) {
        if (blocks[i] !is block && (blocks[i].WaypointType == 'Checkpoint' || blocks[i].WaypointType == 'LinkedCheckpoint')) {
            if (vec3equals(block.position, blocks[i].position)) {
                return true;
            }
        }
    }
    return false;
}

bool HasGateOverlay(Block@ block, Block@[]@ blocks) {
    for (uint i = 0; i < blocks.Length; i++) {
        if (blocks[i] !is block && blocks[i].name.StartsWith('GateSpecial')) {
            if (vec3equals(block.position, blocks[i].position+vec3(0,8,0)) || vec3equals(block.position, blocks[i].position+vec3(0,16,0))) {
                return true;
            }
        }
    }
    return false;
}


bool PoolPlatformOverlay(Block@ block, Block@[]@ blocks) {
    for (uint i = 0; i < blocks.Length; i++) {
        if (blocks[i].iscustom && blocks[i] !is block) {
            if (vec3equals(block.position, blocks[i].position + vec3(0,8,0))) {
                return true;
            }
        }
    }
    return false;
}
bool PoolSceneryOverlay(Block@ block, Block@[]@ blocks) {
    for (uint i = 0; i < blocks.Length; i++) {
        if (blocks[i].iscustom && blocks[i] !is block) {
            if (vec3equals(block.position, blocks[i].position) || (blocks[i].name.Contains('1x3') && ( 
                (blocks[i].direction == CGameCtnBlock::ECardinalDirections::North && (vec3equals(block.position, blocks[i].position + vec3(32, 0, 0)) || vec3equals(block.position, blocks[i].position + vec3(64, 0, 0)))) || 
                (blocks[i].direction == CGameCtnBlock::ECardinalDirections::South && (vec3equals(block.position, blocks[i].position + vec3(-32, 0, 0)) || vec3equals(block.position, blocks[i].position + vec3(-64, 0, 0)))) || 
                (blocks[i].direction == CGameCtnBlock::ECardinalDirections::East && (vec3equals(block.position, blocks[i].position + vec3(0, 0, 32)) || vec3equals(block.position, blocks[i].position + vec3(0, 0, 64)))) || 
                (blocks[i].direction == CGameCtnBlock::ECardinalDirections::West && (vec3equals(block.position, blocks[i].position + vec3(0, 0, -32)) || vec3equals(block.position, blocks[i].position + vec3(0, 0, -64))))))) {
                return true;
            }
        }
    }
    return false;
}

bool StandFinishOverlay(Block@ block, Block@[]@ blocks) {
    return false; // not implemented yet
}

// Assumes return of 1 from ShouldSaveBlock
bool ShouldHighlightBlock(Block@ block, Block@[]@ blocks) {
    if (block.iscustom) return false;
    if (!block.isblock) return ShouldHighlightItem(block, blocks);
    // if (Ground) {
    //     if (block.isground) {
    //         return true;
    //     }
    // }
    if (RoadTech) {
        if (block.name.StartsWith('RoadTech') || block.name.Contains('ToRoadTech')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RoadDirt) {
        if (block.name.StartsWith('RoadDirt') || block.name.Contains('ToRoadDirt')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RallyCastleRoad) {
        if (block.name.StartsWith('RallyCastleRoad')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RallyRoadDirt) {
        if (block.name.StartsWith('RallyRoadDirt')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RoadBump) {
        if (block.name.StartsWith('RoadBump') || block.name.Contains('ToRoadBump')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RoadIce) {
        if (block.name.StartsWith('RoadIce') || block.name.Contains('ToRoadIce')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RoadWood) {
        if (block.name.StartsWith('SnowRoad') || block.name.Contains('ToThemeSnowRoad')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (RoadWater) {
        if (block.name.StartsWith('RoadWater') || block.name.Contains('ToRoadWater')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    // if (TrackWall) {
    //     if (block.name.StartsWith('TrackWall')) {
    //         return true;
    //     }
    // }

    // if (PlatformTech) {
    //     if (block.name.StartsWith('PlatformTech')) {
    //         return true;
    //     }
    // }
    // if (PlatformDirt) {
    //     if (block.name.StartsWith('PlatformDirt')) {
    //         return true;
    //     }
    // }
    // if (PlatformIce) {
    //     if (block.name.StartsWith('PlatformIce')) {
    //         return true;
    //     }
    // }
    // if (PlatformGrass) {
    //     if (block.name.StartsWith('PlatformGrass')) {
    //         return true;
    //     }
    // }
    // if (PlatformPlastic) {
    //     if (block.name.StartsWith('PlatformPlastic')) {
    //         return true;
    //     }
    // }
    // if (PlatformWater) {
    //     if (block.name.StartsWith('PlatformWater')) {
    //         return true;
    //     }
    // }
    // if (DecoWall) {
    //     if (block.name.StartsWith('DecoWall')) {
    //         return true; //return !HasCustomOverlay(block, blocks);
    //     }
    // }

    if (OpenTech) {
        if (block.name.StartsWith('OpenTech')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (OpenDirt) {
        if (block.name.StartsWith('OpenDirt')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (OpenIce) {
        if (block.name.StartsWith('OpenIce')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (OpenGrass) {
        if (block.name.StartsWith('OpenGrass')) {
            if (!HasCustomOverlay(block, blocks)) {
                return true;
            }
        }
    }

    if((CheckpointNoGate || CPfull) && block.WaypointType != '') {// IsCheckpoint
        if (CPfull) {
            // if (block.WaypointType == 'LinkedCheckpoint') {
            //     return 2;
            // }
        } else if (block.WaypointType == 'Checkpoint' || block.WaypointType == 'LinkedCheckpoint') {
            // if (IsCheckpoint) {
            //     return true;
            // }
            if (!HasGateOverlay(block, blocks)) {
                return true;
            }
        }
    } else if (CPfull) {
        // if (HasCheckpoint_Blocks.Find(block.name) >= 0) {
        //     return true;
        // }
        if ((block.name.EndsWith('ToRoadTech') || block.name.EndsWith('ToRoadWater')) && !block.name.StartsWith('TrackWall')) {
            if (!HasCheckpointOverlay(block, blocks)) {
                return true;
            }
        }
    }

    // custom blocks already not highlighted
    // if (Boosters) {
    //     if (EffectBlocks) {
    //         if (block.name.Contains('Special')) {
    //             return true;
    //         }
    //     } else if (block.name.Contains('SpecialTurbo') || block.name.Contains('SpecialBoost')) {
    //         return true;
    //     }
    // } else if (EffectBlocks) {
    //     if (block.name.Contains('Special') && !(block.name.Contains('SpecialTurbo') || block.name.Contains('SpecialBoost'))) {
    //         return true;
    //     }
    // }
    // if (Hole) {
    //     if (Hole_Blocks.Find(block.name) >= 0) {
    //         return true;
    //     }
    // }
    if (PoolHunter) {
        if (block.name.StartsWith('TrackWallWater') || block.name.StartsWith('DecoWallWater') || block.name.StartsWith('Water')) {
            if (!PoolSceneryOverlay(block, blocks)) {
                return true;
            }
        } else if (block.name.StartsWith('PlatformWater') || block.name.StartsWith('RoadWater')) {
            if (!PoolPlatformOverlay(block, blocks)) {
                return true;
            }
        }
    }
    if (Manslaughter) {
        if (block.name.StartsWith('Stand')) {
            if (!StandFinishOverlay(block, blocks)) {
                return true;
            }
        }
    }
    return false;
}

bool ShouldHighlightItem(Block@ item, Block@[]@ blocks) {
    // if (ItemPlastic) {
    //     if (item.name.StartsWith('Inflatable')) {
    //         return true;
    //     }
    // }
    if (ItemDirt) {
        if (item.name.EndsWith('Battlements')) {
            if (!HasCustomOverlay(item, blocks)) {
                return true;
            }
        }
    }
    if (ItemWood) {
        if (item.name.StartsWith('SnowTemple') || item.name.StartsWith('SnowBridge')) {
            if (!HasCustomOverlay(item, blocks)) {
                return true;
            }
        }
    }
    if((CheckpointNoGate) && item.WaypointType != '') {// IsCheckpoint || CPfull
        if (CPfull) {
        //     if (item.WaypointSpecialProperty.Tag == 'LinkedCheckpoint') {
        //         return 2;
        //     }
        } else if (item.WaypointType == 'Checkpoint' || item.WaypointType == 'LinkedCheckpoint') {
            // if (IsCheckpoint) {
            //     return true;
            // }
            return true; // no gate check for items yet
        }
    }
    if (Boosters) {
        if (item.name.StartsWith('GateSpecial')) {
            if (EffectBlocks) {
                return true;
            } else if (item.name.Contains('Turbo') || item.name.Contains('Boost')) {
                return true;
            }
        }
    } else if (EffectBlocks) {
        if (item.name.StartsWith('GateSpecial') && !(item.name.Contains('Turbo') || item.name.Contains('Boost'))) {
            return true;
        }
    }
    return false;
}

string[] HasCheckpoint_Blocks = {'RoadTechStraight', 'RoadTechSlopeStraight', 'RoadTechTiltStraight', 'RoadTechDiagLeftStraightX2', 'RoadTechDiagRightStraightX2', 'RoadDirtStraight', 'RoadDirtSlopeStraight', 'RoadDirtTiltStraight', 'RoadDirtDiagLeftStraightX2', 'RoadDirtDiagRightStraightX2', 'RoadBumpStraight', 'RoadBumpSlopeStraight', 'RoadBumpTiltStraight', 'RoadBumpDiagLeftStraightX2', 'RoadBumpDiagRightStraightX2', 'RoadIceStraight', 'RoadIceSlopeStraight', 'RoadIceWithWallStraight', 'RoadIceDiagLeftWithWallStraight', 'RoadIceDiagRightWithWallStraight',
'PlatformTechBase', 'PlatformTechSlope2Straight', 'PlatformTechWallStraight4', 'PlatformDirtBase', 'PlatformDirtSlope2Straight', 'PlatformDirtWallStraight4', 'PlatformIceBase', 'PlatformIceSlope2Straight', 'PlatformIceWallStraight4', 'PlatformGrassBase', 'PlatformGrassSlope2Straight', 'PlatformGrassWallStraight4', 'PlatformPlasticSlope2Straight', 'PlatformPlasticBase', 'PlatformPlasticWallStraight4', 'RoadWaterStraight', 'PlatformWaterRampBase', 'PlatformWaterBase',
'OpenTechRoadStraight', 'OpenTechRoadSlope2Straight', 'OpenTechRoadStraightTilt2', 'OpenDirtRoadStraight', 'OpenDirtRoadSlope2Straight', 'OpenDirtRoadStraightTilt2', 'OpenIceRoadStraight', 'OpenIceRoadSlope2Straight', 'OpenIceRoadStraightTilt2', 'OpenGrassRoadStraight', 'OpenGrassRoadSlope2Straight', 'OpenGrassRoadStraightTilt2', 'OpenPlasticRoadStraight', 'OpenPlasticRoadSlope2Straight', 'OpenPlasticRoadStraightTilt2', 
'SnowRoadStraight', 'RallyRoadDirtLowStraight', 'RallyRoadDirtHighStraight', 'RallyCastleRoadStraight' // gamepadeditor macroblock checkpoints
};


string[] Hole_Blocks = {'RoadTechStraight', 'RoadTechSlopeStraight', 'RoadTechTiltStraight', 'RoadTechDiagLeftStraightX2', 'RoadTechDiagRightStraightX2', 'RoadDirtStraight', 'RoadDirtSlopeStraight', 'RoadDirtTiltStraight', 'RoadDirtDiagLeftStraightX2', 'RoadDirtDiagRightStraightX2', 'RoadBumpStraight', 'RoadBumpSlopeStraight', 'RoadBumpTiltStraight', 'RoadBumpDiagLeftStraightX2', 'RoadBumpDiagRightStraightX2', 'RoadIceStraight', 'RoadIceSlopeStraight', 'RoadIceWithWallStraight', 'RoadIceDiagLeftWithWallStraight', 'RoadIceDiagRightWithWallStraight', 'PlatformTechBase', 'PlatformTechSlope2Straight', 'PlatformTechWallStraight4', 'PlatformDirtBase', 'PlatformDirtSlope2Straight', 'PlatformDirtWallStraight4', 'PlatformIceBase', 'PlatformIceSlope2Straight', 'PlatformIceWallStraight4', 'PlatformGrassBase', 'PlatformGrassSlope2Straight', 'PlatformGrassWallStraight4', 'PlatformPlasticBase', 'PlatformPlasticSlope2Straight', 'PlatformPlasticWallStraight4', 'RoadWaterStraight', 'TrackWallWaterStraight', 'TrackWallSlopeStraight', 'TrackWallStraight', 'TrackWallTiltStraight', 'TrackWallDiagLeftStraightX2', 'TrackWallDiagRightStraightX2'};
