
[Setting name="Auto Color Selection"]
bool autoColor = true;

[Setting name="Checkpoint Quantity Warning"]
bool cpWarning = true;

[Setting name="Color plugins menu name"]
bool colorDisplayName = true;

[Setting name="Check overlay hitbox"]
bool checkOverlayHitbox = false;

[Setting name="Show Debug position dots"]
bool showDebug = false;

enum VariantColorOptions {
    Disabled,
    Ghost,
    NotGhost,
    Ground,
    NotGround,
    Free,
    NotFree,
    Normal,
    NotNormal

}
[Setting name="Variant Color"]
VariantColorOptions variantColor = VariantColorOptions::NotNormal;

string VariantColorName (VariantColorOptions v) {
    switch (v) {
        case (VariantColorOptions::Disabled): return 'Disabled';
        case (VariantColorOptions::Ghost): return 'Ghost blocks';
        case (VariantColorOptions::NotGhost): return 'Non-ghost blocks';
        case (VariantColorOptions::Free): return 'Freeblocked blocks';
        case (VariantColorOptions::NotFree): return 'Non-freeblocked blocks';
        case (VariantColorOptions::Ground): return 'Ground mode blocks';
        case (VariantColorOptions::NotGround): return 'Air mode blocks';
        case (VariantColorOptions::NotNormal): return 'Ghost or freeblock mode blocks';
        case (VariantColorOptions::Normal): return 'Normal mode blocks';
        default: return 'Unknown';
    }
}

int3 mapId = int3(-1, -1, -1);

const string filename = IO::FromStorageFolder('CPData.json');

void onOpenEditor() {
    ineditor = true;
    CGameCtnApp@ app = GetApp();
    CGameCtnChallenge@ rootmap = app.RootMap;
    CGameCtnEditorFree@ editor = cast<CGameCtnEditorFree>(GetApp().Editor);
    CGameEditorPluginMapMapType@ pluginmap = editor.PluginMapType;

    string mapname = rootmap.MapName;
    int i = -1;
    if ((i = mapname.IndexOf('Winter')) >= 0) {
        i += 'Winter'.Length + 1;
        mapId.y = 3;
    } else if ((i = mapname.IndexOf('Spring')) >= 0) {
        i += 'Spring'.Length + 1;
        mapId.y = 7;
    } else if ((i = mapname.IndexOf('Summer')) >= 0) {
        i += 'Summer'.Length + 1;
        mapId.y = 11;
    } else if ((i = mapname.IndexOf('Fall')) >= 0) {
        i += 'Fall'.Length + 1;
        mapId.y = 15;
    }
    if (i >= 0) {
        mapId.x = Text::ParseInt(mapname.SubStr(i, 4));
        i += 4+3;
        mapId.z = Text::ParseInt(mapname.SubStr(i, 2));
        if (mapId.z == 0) {
            mapId = int3(-1, -1, -1);
            return;} // season too close to end of name
        if (autoColor && (mapId.x > 2021 || (mapId.x == 2021 && mapId.y > 7))) {
            if (mapId.z <= 5) {
                pluginmap.NextMapElemColor = CGameEditorPluginMap::EMapElemColor::White;
            } else if (mapId.z <= 10) {
                pluginmap.NextMapElemColor = CGameEditorPluginMap::EMapElemColor::Green;
            } else if (mapId.z <= 15) {
                pluginmap.NextMapElemColor = CGameEditorPluginMap::EMapElemColor::Blue;
            } else if (mapId.z <= 20) {
                pluginmap.NextMapElemColor = CGameEditorPluginMap::EMapElemColor::Red;
            } else if (mapId.z <= 25) {
                pluginmap.NextMapElemColor = CGameEditorPluginMap::EMapElemColor::Black;
            }
        }
        if (mapId.x >= 2020 && getCpCount() == -1) { // checking > 2020 since I know lower years won't have this
            startnew(updateCPsFromSite);
        }
    } else {
        mapId = int3(-1, -1, -1);
    }
}

void updateCPsFromSite() {
    trace('unknown map, requesting new cp data from Openplanet website');
    Net::HttpRequest@ getCPs = Net::HttpRequest();
    getCPs.Url = "https://openplanet.dev/plugin/alteredcampaignhelper/config/cpdata";
    getCPs.Start();
    while (!getCPs.Finished()) {
        yield();
    }
    if (getCPs.ResponseCode() == 200) {
        trace('request ok');
    } else {
        trace('request failed');
        UI::ShowNotification("Altered Campaign Helper", "Failed to connect to server", vec4(0.5,0,0,0.5));
    }
    getCPs.SaveToFile(filename);
    if (ineditor) {
        if (getCpCount() >= 0) {
            trace('new cp data found for map');
            UI::ShowNotification("Altered Campaign Helper", "Downloaded new cp data from server", vec4(0,0.5,0,0.5));
        } else {
            trace('still no cp data for this map');
        }
    } else {
        trace('map exited before new cp data check');
    }
}


void checkCheckpoints() {
    if (cpWarning) {
        int cpCount;
        int originalCpCount = getCpCount();
        if (originalCpCount != -1 && originalCpCount != (cpCount = countCheckpoints())) {
            UI::ShowNotification("Altered Campaign Helper", "Checkpoint count " + cpCount + " is different to original, " + originalCpCount, vec4(0.5,0,0,0.5));
        }// else {
        //    UI::ShowNotification("Altered Campaign Helper", "Checkpoint count " + cpCount + " matches", vec4(0,0.5,0,0.5));
        // }
    }
}

void onCloseEditor() {
    ineditor = false;
}

// gets the CP count expected for the map name
int getCpCount() {
    Json::Value JsonCps = Json::FromFile(filename);
    if (JsonCps.GetType() == Json::Type::Null || !JsonCps.HasKey(tostring(mapId.x)) || !JsonCps[tostring(mapId.x)].HasKey(tostring(mapId.y)) || !JsonCps[tostring(mapId.x)][tostring(mapId.y)].HasKey(tostring(mapId.y))) {
        return -1;
    } else {
        return Text::ParseInt(JsonCps[tostring(mapId.x)][tostring(mapId.y)][tostring(mapId.z)]);
    }
}

// Counts the checkpoints in the current map
int countCheckpoints () {
    CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);
    MwFastBuffer<CGameScriptMapLandmark@> landmarks = playground.Arena.MapLandmarks;
    int[] CPorders = {};
    int cps = 0;
    for(uint i = 0; i < landmarks.Length; i++) {

        if ((landmarks[i].Waypoint is null)) {
        } else {
            if (landmarks[i].Tag == 'Checkpoint') {
                // print('cp: ' + landmarks[i].Order);
                cps++;
            } else if (landmarks[i].Tag == 'LinkedCheckpoint') {
                if (CPorders.Find(landmarks[i].Order) < 0) {
                    // print('lcp: ' + landmarks[i].Order);
                    cps++;
                }
                CPorders.InsertLast(landmarks[i].Order);
            }
        }
    }
    // print(cps);
    return cps;
}

#if GET_CP_DATA
// Idk if this code is needed in published version, but might be useful for someone
Json::Value cpCounts = Json::Object();

void addCpCount() {
    CGameCtnApp@ app = GetApp();
    CGameCtnChallenge@ rootmap = app.RootMap;

    string mapname = rootmap.MapName;
    string season;
    int i = -1;
    if ((i = mapname.IndexOf('Winter')) >= 0) {
        season = 'Winter';
        mapId.y = 3;
    } else if ((i = mapname.IndexOf('Spring')) >= 0) {
        season = 'Spring';
        mapId.y = 7;
    } else if ((i = mapname.IndexOf('Summer')) >= 0) {
        season = 'Summer';
        mapId.y = 11;
    } else if ((i = mapname.IndexOf('Fall')) >= 0) {
        season = 'Fall';
        mapId.y = 15;
    }
    if (i >= 0) {
        i += season.Length + 1;
        mapId.x = Text::ParseInt(mapname.SubStr(i, 4));
        i += 4+3;
        mapId.z = Text::ParseInt(mapname.SubStr(i, 2));
    }

    if (cpCounts[tostring(mapId.x)].GetType() == 0) {
        cpCounts[tostring(mapId.x)] = Json::Object();
    }
    if (cpCounts[tostring(mapId.x)][tostring(mapId.y)].GetType() == 0) {
        cpCounts[tostring(mapId.x)][tostring(mapId.y)] = Json::Object();
    }

    if (cpCounts[tostring(mapId.x)].GetType() == 0) {
        cpCounts[tostring(mapId.x)] = Json::Object();
    }
    if (cpCounts[tostring(mapId.x)][tostring(mapId.y)].GetType() == 0) {
        cpCounts[tostring(mapId.x)][tostring(mapId.y)] = Json::Object();
    }
    cpCounts[tostring(mapId.x)][tostring(mapId.y)][tostring(mapId.z)] = tostring(countCheckpoints());
    print(cpCounts[tostring(mapId.x)][tostring(mapId.y)][tostring(mapId.z)]);
}

void saveCpCounts() {
    Json::ToFile(IO::FromStorageFolder('Plugins/AlteredCampaignHelper/CPs_saved.json'), cpCounts); // saves in storage folder
}
#endif