// mostly from XertoV's Item Placement Toolbox (which is 'unlicense')

const uint16 FreeBlockPosOffset = Reflection::GetType("CGameCtnBlock").GetMember("Dir").Offset + 0x8;
const uint16 FreeBlockRotOffset = FreeBlockPosOffset + 0xC;

vec3 GetBlockLocation(CGameCtnBlock@ block) {
    if (IsBlockFree(block)) {
        // free block mode
        return Dev::GetOffsetVec3(block, FreeBlockPosOffset);
    }
    // using the coord will not give you a consistent corner of the block (i.e., after rotation),
    // pre-adjust the coordinates to account for this based on cardinal dir
    // rclick in editor always rotates around BL (min x/z);
    vec3 coord = Nat3ToVec3(block.Coord);
    vec3 coordSize = GetBlockCoordSize(block);
    if (int(block.Dir) == 1) {
        coord.x += coordSize.z - 1;
    }
    if (int(block.Dir) == 2) {
        coord.x += coordSize.x - 1;
        coord.z += coordSize.z - 1;
    }
    if (int(block.Dir) == 3) {
        coord.z += coordSize.x - 1;
    }
    vec3 pos = CoordToPos(coord);
    vec3 sqSize = vec3(32, 8, 32);
    vec3 rot = GetBlockRotation(block);
    return (mat4::Translate(pos) * mat4::Translate(sqSize / 2.) * EulerToMat(rot) * (sqSize / -2.)).xyz;
}

vec3 GetBlockRotation(CGameCtnBlock@ block) {
    if (IsBlockFree(block)) {
        // free block mode
        vec3 ypr = Dev::GetOffsetVec3(block, FreeBlockRotOffset);
        return vec3(ypr.y, ypr.x, ypr.z);
    }
    // n:0, e:1, s:2, w:3
    return vec3(0,  CardinalDirectionToYaw(int(block.Dir)), 0);
}

float CardinalDirectionToYaw(int dir) {
    // n:0, e:1, s:2, w:3
    return -Math::PI/2. * float(dir) + (dir >= 2 ? Math::PI*2 : 0);
}

bool IsBlockFree(CGameCtnBlock@ block) {
    return int(block.CoordX) < 0;
}

vec3 Nat3ToVec3(nat3 coord) {
    return vec3(coord.x, coord.y, coord.z);
}

vec3 CoordToPos(vec3 coord) {
    return vec3(coord.x * 32, (int(coord.y) - 8) * 8, coord.z * 32);
} vec3 CoordToPos(nat3 coord) {
    return vec3(coord.x * 32, (int(coord.y) - 8) * 8, coord.z * 32);
}

vec3 GetBlockSize(CGameCtnBlock@ block) {
    return GetBlockCoordSize(block) * vec3(32, 2, 32);
}

vec3 GetBlockCoordSize(CGameCtnBlock@ block) {
    CGameCtnBlockInfoVariant@ biv = GetBlockInfoVariant(block);
    return Nat3ToVec3(biv.Size);
}

CGameCtnBlockInfoVariant@ GetBlockInfoVariant(CGameCtnBlock@ block) {
    uint bivIx = block.BlockInfoVariantIndex;
    CGameCtnBlockInfo@ bi = block.BlockInfo;
    CGameCtnBlockInfoVariant@ biv;
    if (bivIx > 0) {
        @biv = block.IsGround ? cast<CGameCtnBlockInfoVariant>(bi.AdditionalVariantsGround[bivIx - 1]) : cast<CGameCtnBlockInfoVariant>(bi.AdditionalVariantsAir[bivIx - 1]);
    } else {
        @biv = block.IsGround ? cast<CGameCtnBlockInfoVariant>(bi.VariantGround) : cast<CGameCtnBlockInfoVariant>(bi.VariantAir);
        if (biv is null) {
            @biv = block.IsGround ? cast<CGameCtnBlockInfoVariant>(bi.VariantBaseGround) : cast<CGameCtnBlockInfoVariant>(bi.VariantBaseAir);
        }
    }
    return biv;
}

vec3 GetCtnBlockMidpoint(CGameCtnBlock@ block) {
    return (GetBlockMatrix(block) * (GetBlockSize(block) / 2)).xyz;
}
vec3 GetCtnBlockPosition(CGameCtnBlock@ block) {
    if (IsBlockFree(block)) {
        // free block mode
        return Dev::GetOffsetVec3(block, FreeBlockPosOffset);
    }
    // for comparing if 2 blocks are in the same space, even if they have different rotation.
    return CoordToPos(block.Coord);
}


mat4 GetBlockMatrix(CGameCtnBlock@ block) {
    return mat4::Translate(GetBlockLocation(block)) * EulerToMat(GetBlockRotation(block));
}


// Use it like auto m = mat4::Translate(pos) * EulerToMat(rot);  with:
// From Rxelux's `mat4x` lib, modified
mat4 EulerToMat(vec3 euler) {
    // mat4 translation = mat4::Translate(position*-1);
    mat4 pitch = mat4::Rotate(-euler.x,vec3(1,0,0));
    mat4 yaw = mat4::Rotate(-euler.y,vec3(0,1,0));
    mat4 roll = mat4::Rotate(-euler.z,vec3(0,0,1));
    return mat4::Inverse(pitch*roll*yaw/* *translation */);
}